@@ -5,13 +5,13 @@
 package org.mozilla.geckoview;
 
 import android.annotation.SuppressLint;
-import android.os.Build;
 import android.util.Log;
 import android.util.SparseArray;
 import androidx.annotation.AnyThread;
 import androidx.annotation.IntDef;
 import androidx.annotation.NonNull;
 import androidx.annotation.Nullable;
+import androidx.annotation.StringDef;
 import androidx.annotation.UiThread;
 import java.lang.annotation.Retention;
 import java.lang.annotation.RetentionPolicy;
@@ -87,17 +87,14 @@ public class WebExtensionController {
       final GeckoBundle bundle = new GeckoBundle(1);
       bundle.putString("extensionId", id);
 
-      final GeckoResult<WebExtension> pending =
-          EventDispatcher.getInstance()
-              .queryBundle("GeckoView:WebExtension:Get", bundle)
-              .map(
-                  extensionBundle -> {
-                    final WebExtension ext = mObserver.onNewExtension(extensionBundle);
-                    mData.put(ext.id, ext);
-                    return ext;
-                  });
-
-      return pending;
+      return EventDispatcher.getInstance()
+          .queryBundle("GeckoView:WebExtension:Get", bundle)
+          .map(
+              extensionBundle -> {
+                final WebExtension ext = mObserver.onNewExtension(extensionBundle);
+                mData.put(ext.id, ext);
+                return ext;
+              });
     }
 
     public void setObserver(final Observer observer) {
@@ -250,6 +247,31 @@ public class WebExtensionController {
    */
   @UiThread
   public interface PromptDelegate {
+
+    /**
+     * Called whenever a new extension is being installed. This is intended as an opportunity for
+     * the app to prompt the user for the permissions required by this extension.
+     *
+     * @deprecated Use onInstallPromptRequest(WebExtension, String[], String[], String[]) to account
+     *     for data collection permissions.
+     * @param extension The {@link WebExtension} that is about to be installed. You can use {@link
+     *     WebExtension#metaData} to gather information about this extension when building the user
+     *     prompt dialog.
+     * @param permissions The list of permissions that are granted during installation.
+     * @param origins The list of origins that are granted during installation.
+     * @return A {@link GeckoResult} that completes with a {@link
+     *     WebExtension.PermissionPromptResponse} containing all the details from the user response.
+     */
+    @Nullable
+    @Deprecated
+    @DeprecationSchedule(id = "web-extension-on-install-prompt-request", version = 143)
+    default GeckoResult<WebExtension.PermissionPromptResponse> onInstallPromptRequest(
+        @NonNull final WebExtension extension,
+        @NonNull final String[] permissions,
+        @NonNull final String[] origins) {
+      return null;
+    }
+
     /**
      * Called whenever a new extension is being installed. This is intended as an opportunity for
      * the app to prompt the user for the permissions required by this extension.
@@ -257,13 +279,19 @@ public class WebExtensionController {
      * @param extension The {@link WebExtension} that is about to be installed. You can use {@link
      *     WebExtension#metaData} to gather information about this extension when building the user
      *     prompt dialog.
-     * @return A {@link GeckoResult} that completes to either {@link AllowOrDeny#ALLOW ALLOW} if
-     *     this extension should be installed or {@link AllowOrDeny#DENY DENY} if this extension
-     *     should not be installed. A null value will be interpreted as {@link AllowOrDeny#DENY
-     *     DENY}.
+     * @param permissions The list of permissions that are granted during installation.
+     * @param origins The list of origins that are granted during installation.
+     * @param dataCollectionPermissions The list of data collection permissions that are requested
+     *     during installation.
+     * @return A {@link GeckoResult} that completes with a {@link
+     *     WebExtension.PermissionPromptResponse} containing all the details from the user response.
      */
     @Nullable
-    default GeckoResult<AllowOrDeny> onInstallPrompt(final @NonNull WebExtension extension) {
+    default GeckoResult<WebExtension.PermissionPromptResponse> onInstallPromptRequest(
+        @NonNull final WebExtension extension,
+        @NonNull final String[] permissions,
+        @NonNull final String[] origins,
+        @NonNull final String[] dataCollectionPermissions) {
       return null;
     }
 
@@ -271,6 +299,8 @@ public class WebExtensionController {
      * Called whenever an updated extension has new permissions. This is intended as an opportunity
      * for the app to prompt the user for the new permissions required by this extension.
      *
+     * @deprecated Use onUpdatePrompt(WebExtension, String[], String[], String[]) to account for
+     *     data collection permissions.
      * @param currentlyInstalled The {@link WebExtension} that is currently installed.
      * @param updatedExtension The {@link WebExtension} that will replace the previous extension.
      * @param newPermissions The new permissions that are needed.
@@ -280,6 +310,8 @@ public class WebExtensionController {
      *     not be update. A null value will be interpreted as {@link AllowOrDeny#DENY DENY}.
      */
     @Nullable
+    @Deprecated
+    @DeprecationSchedule(id = "web-extension-on-update-prompt", version = 144)
     default GeckoResult<AllowOrDeny> onUpdatePrompt(
         @NonNull final WebExtension currentlyInstalled,
         @NonNull final WebExtension updatedExtension,
@@ -288,6 +320,52 @@ public class WebExtensionController {
       return null;
     }
 
+    /**
+     * Called whenever an updated extension has new permissions. This is intended as an opportunity
+     * for the app to prompt the user for the new permissions required by this extension.
+     *
+     * @param extension The {@link WebExtension} being updated.
+     * @param newPermissions The new permissions that are needed.
+     * @param newOrigins The new origins that are needed.
+     * @param newDataCollectionPermissions The new data collection permissions that are needed.
+     * @return A {@link GeckoResult} that completes to either {@link AllowOrDeny#ALLOW ALLOW} if
+     *     this extension should be update or {@link AllowOrDeny#DENY DENY} if this extension should
+     *     not be update. A null value will be interpreted as {@link AllowOrDeny#DENY DENY}.
+     */
+    @Nullable
+    default GeckoResult<AllowOrDeny> onUpdatePrompt(
+        @NonNull final WebExtension extension,
+        @NonNull final String[] newPermissions,
+        @NonNull final String[] newOrigins,
+        @NonNull final String[] newDataCollectionPermissions) {
+      return null;
+    }
+
+    /**
+     * Called whenever permissions are requested. This is intended as an opportunity for the app to
+     * prompt the user for the permissions required by this extension at runtime.
+     *
+     * @deprecated Use onOptionalPrompt(WebExtension, String[], String[], String[]) to account for
+     *     data collection permissions.
+     * @param extension The {@link WebExtension} that is about to be installed. You can use {@link
+     *     WebExtension#metaData} to gather information about this extension when building the user
+     *     prompt dialog.
+     * @param permissions The permissions that are requested.
+     * @param origins The requested host permissions.
+     * @return A {@link GeckoResult} that completes to either {@link AllowOrDeny#ALLOW ALLOW} if the
+     *     request should be approved or {@link AllowOrDeny#DENY DENY} if the request should be
+     *     denied. A null value will be interpreted as {@link AllowOrDeny#DENY DENY}.
+     */
+    @Nullable
+    @Deprecated
+    @DeprecationSchedule(id = "web-extension-on-optional-prompt", version = 144)
+    default GeckoResult<AllowOrDeny> onOptionalPrompt(
+        final @NonNull WebExtension extension,
+        final @NonNull String[] permissions,
+        final @NonNull String[] origins) {
+      return null;
+    }
+
     /**
      * Called whenever permissions are requested. This is intended as an opportunity for the app to
      * prompt the user for the permissions required by this extension at runtime.
@@ -297,6 +375,7 @@ public class WebExtensionController {
      *     prompt dialog.
      * @param permissions The permissions that are requested.
      * @param origins The requested host permissions.
+     * @param dataCollectionPermissions The requested data collection permissions.
      * @return A {@link GeckoResult} that completes to either {@link AllowOrDeny#ALLOW ALLOW} if the
      *     request should be approved or {@link AllowOrDeny#DENY DENY} if the request should be
      *     denied. A null value will be interpreted as {@link AllowOrDeny#DENY DENY}.
@@ -305,7 +384,8 @@ public class WebExtensionController {
     default GeckoResult<AllowOrDeny> onOptionalPrompt(
         final @NonNull WebExtension extension,
         final @NonNull String[] permissions,
-        final @NonNull String[] origins) {
+        final @NonNull String[] origins,
+        final @NonNull String[] dataCollectionPermissions) {
       return null;
     }
   }
@@ -338,6 +418,14 @@ public class WebExtensionController {
     @UiThread
     default void onDisabling(@NonNull WebExtension extension) {}
 
+    /**
+     * Called whenever optional permissions of an extension have changed.
+     *
+     * @param extension The {@link WebExtension} that has optional permissions changed.
+     */
+    @UiThread
+    default void onOptionalPermissionsChanged(@NonNull WebExtension extension) {}
+
     /**
      * Called whenever an extension has been disabled.
      *
@@ -403,6 +491,16 @@ public class WebExtensionController {
     @UiThread
     default void onInstallationFailed(
         final @Nullable WebExtension extension, final @NonNull InstallException installException) {}
+
+    /**
+     * Called whenever an extension startup has been completed (and relative urls to assets packaged
+     * with the extension can be resolved into a full moz-extension url, e.g. optionsPageUrl is
+     * going to be empty until the extension has reached this callback).
+     *
+     * @param extension The {@link WebExtension} that has been fully started.
+     */
+    @UiThread
+    default void onReady(final @NonNull WebExtension extension) {}
   }
 
   /** This delegate is used to notify of extension process state changes. */
@@ -482,6 +580,7 @@ public class WebExtensionController {
       EventDispatcher.getInstance()
           .unregisterUiThreadListener(
               mInternals,
+              "GeckoView:WebExtension:OnOptionalPermissionsChanged",
               "GeckoView:WebExtension:OnDisabling",
               "GeckoView:WebExtension:OnDisabled",
               "GeckoView:WebExtension:OnEnabling",
@@ -490,11 +589,13 @@ public class WebExtensionController {
               "GeckoView:WebExtension:OnUninstalled",
               "GeckoView:WebExtension:OnInstalling",
               "GeckoView:WebExtension:OnInstallationFailed",
-              "GeckoView:WebExtension:OnInstalled");
+              "GeckoView:WebExtension:OnInstalled",
+              "GeckoView:WebExtension:OnReady");
     } else if (delegate != null && mAddonManagerDelegate == null) {
       EventDispatcher.getInstance()
           .registerUiThreadListener(
               mInternals,
+              "GeckoView:WebExtension:OnOptionalPermissionsChanged",
               "GeckoView:WebExtension:OnDisabling",
               "GeckoView:WebExtension:OnDisabled",
               "GeckoView:WebExtension:OnEnabling",
@@ -503,7 +604,8 @@ public class WebExtensionController {
               "GeckoView:WebExtension:OnUninstalled",
               "GeckoView:WebExtension:OnInstalling",
               "GeckoView:WebExtension:OnInstallationFailed",
-              "GeckoView:WebExtension:OnInstalled");
+              "GeckoView:WebExtension:OnInstalled",
+              "GeckoView:WebExtension:OnReady");
     }
 
     mAddonManagerDelegate = delegate;
@@ -545,6 +647,21 @@ public class WebExtensionController {
     EventDispatcher.getInstance().dispatch("GeckoView:WebExtension:EnableProcessSpawning", null);
   }
 
+  /**
+   * Disable extension process spawning.
+   *
+   * <p>Extension process spawning can be re-enabled with {@link
+   * WebExtensionController#enableExtensionProcessSpawning()}. This method does the opposite and
+   * stops the extension process. This method can be called when we no longer want to run extensions
+   * for the rest of the session.
+   *
+   * @see ExtensionProcessDelegate#onDisabledProcessSpawning()
+   */
+  @AnyThread
+  public void disableExtensionProcessSpawning() {
+    EventDispatcher.getInstance().dispatch("GeckoView:WebExtension:DisableProcessSpawning", null);
+  }
+
   private static class InstallCanceller implements GeckoResult.CancellationDelegate {
     public final String installId;
 
@@ -575,7 +692,61 @@ public class WebExtensionController {
    *
    * <p>When calling this method, the GeckoView library will download the extension, validate its
    * manifest and signature, and give you an opportunity to verify its permissions through {@link
-   * PromptDelegate#installPrompt}, you can use this method to prompt the user if appropriate.
+   * PromptDelegate#onInstallPromptRequest}, you can use this method to prompt the user if
+   * appropriate.
+   *
+   * @param uri URI to the extension's <code>.xpi</code> package. This can be a remote <code>https:
+   *     </code> URI or a local <code>file:</code> or <code>resource:</code> URI. Note: the app
+   *     needs the appropriate permissions for local URIs.
+   * @param installationMethod The method used by the embedder to install the {@link WebExtension}.
+   * @return A {@link GeckoResult} that will complete when the installation process finishes. For
+   *     successful installations, the GeckoResult will return the {@link WebExtension} object that
+   *     you can use to set delegates and retrieve information about the WebExtension using {@link
+   *     WebExtension#metaData}.
+   *     <p>If an error occurs during the installation process, the GeckoResult will complete
+   *     exceptionally with a {@link WebExtension.InstallException InstallException} that will
+   *     contain the relevant error code in {@link WebExtension.InstallException#code
+   *     InstallException#code}.
+   * @see PromptDelegate#onInstallPromptRequest(WebExtension, String[], String[], String[])
+   * @see WebExtension.InstallException.ErrorCodes
+   * @see WebExtension#metaData
+   */
+  @NonNull
+  @AnyThread
+  public GeckoResult<WebExtension> install(
+      final @NonNull String uri, final @Nullable @InstallationMethod String installationMethod) {
+    final InstallCanceller canceller = new InstallCanceller();
+    final GeckoBundle bundle = new GeckoBundle(3);
+    bundle.putString("locationUri", uri);
+    bundle.putString("installId", canceller.installId);
+    bundle.putString("installMethod", installationMethod);
+
+    final GeckoResult<WebExtension> result =
+        EventDispatcher.getInstance()
+            .queryBundle("GeckoView:WebExtension:Install", bundle)
+            .map(
+                ext -> WebExtension.fromBundle(mDelegateControllerProvider, ext),
+                WebExtension.InstallException::fromQueryException)
+            .map(this::registerWebExtension);
+    result.setCancellationDelegate(canceller);
+    return result;
+  }
+
+  /**
+   * Install an extension.
+   *
+   * <p>An installed extension will persist and will be available even when restarting the {@link
+   * GeckoRuntime}.
+   *
+   * <p>Installed extensions through this method need to be signed by Mozilla, see <a
+   * href="https://extensionworkshop.com/documentation/publish/signing-and-distribution-overview/#distributing-your-addon">
+   * Distributing your add-on </a>.
+   *
+   * <p>When calling this method, the GeckoView library will download the extension, validate its
+   * manifest and signature, and give you an opportunity to verify its permissions through {@link
+   * PromptDelegate#installPromptRequest(GeckoBundle, EventCallback)}, you can use this method to
+   * prompt the user if appropriate. If you are looking to provide an {@link InstallationMethod},
+   * please use {@link WebExtensionController#install(String, String)}
    *
    * @param uri URI to the extension's <code>.xpi</code> package. This can be a remote <code>https:
    *     </code> URI or a local <code>file:</code> or <code>resource:</code> URI. Note: the app
@@ -588,29 +759,34 @@ public class WebExtensionController {
    *     exceptionally with a {@link WebExtension.InstallException InstallException} that will
    *     contain the relevant error code in {@link WebExtension.InstallException#code
    *     InstallException#code}.
-   * @see PromptDelegate#installPrompt
+   * @see PromptDelegate#installPromptRequest
    * @see WebExtension.InstallException.ErrorCodes
    * @see WebExtension#metaData
    */
   @NonNull
   @AnyThread
   public GeckoResult<WebExtension> install(final @NonNull String uri) {
-    final InstallCanceller canceller = new InstallCanceller();
-    final GeckoBundle bundle = new GeckoBundle(2);
-    bundle.putString("locationUri", uri);
-    bundle.putString("installId", canceller.installId);
-
-    final GeckoResult<WebExtension> result =
-        EventDispatcher.getInstance()
-            .queryBundle("GeckoView:WebExtension:Install", bundle)
-            .map(
-                ext -> WebExtension.fromBundle(mDelegateControllerProvider, ext),
-                WebExtension.InstallException::fromQueryException)
-            .map(this::registerWebExtension);
-    result.setCancellationDelegate(canceller);
-    return result;
+    return install(uri, null);
   }
 
+  /** The method used by the embedder to install the {@link WebExtension}. */
+  @Retention(RetentionPolicy.SOURCE)
+  @StringDef({
+    INSTALLATION_METHOD_MANAGER,
+    INSTALLATION_METHOD_FROM_FILE,
+    INSTALLATION_METHOD_ONBOARDING
+  })
+  public @interface InstallationMethod {};
+
+  /** Indicates the {@link WebExtension} was installed using from the embedder's add-ons manager. */
+  public static final String INSTALLATION_METHOD_MANAGER = "manager";
+
+  /** Indicates the {@link WebExtension} was installed from a file. */
+  public static final String INSTALLATION_METHOD_FROM_FILE = "install-from-file";
+
+  /** Indicates the {@link WebExtension} was installed from the embedder's onboarding feature. */
+  public static final String INSTALLATION_METHOD_ONBOARDING = "onboarding";
+
   /**
    * Set whether an extension should be allowed to run in private browsing or not.
    *
@@ -633,6 +809,106 @@ public class WebExtensionController {
         .map(this::registerWebExtension);
   }
 
+  /**
+   * Add the provided permissions to the {@link WebExtension} with the given id.
+   *
+   * @deprecated Use addOptionalPermissions(WebExtension, String[], String[], String[]) to account
+   *     for data collection permissions.
+   * @param extensionId the id of {@link WebExtension} instance to modify.
+   * @param permissions the permissions to add, pass an empty array to not update.
+   * @param origins the origins to add, pass an empty array to not update.
+   * @return the updated {@link WebExtension} instance.
+   */
+  @NonNull
+  @AnyThread
+  @Deprecated
+  @DeprecationSchedule(id = "web-extension-add-optional-permissions", version = 143)
+  public GeckoResult<WebExtension> addOptionalPermissions(
+      final @NonNull String extensionId,
+      @NonNull final String[] permissions,
+      @NonNull final String[] origins) {
+    return addOptionalPermissions(extensionId, permissions, origins, new String[0]);
+  }
+
+  /**
+   * Add the provided permissions to the {@link WebExtension} with the given id.
+   *
+   * @param extensionId the id of {@link WebExtension} instance to modify.
+   * @param permissions the permissions to add, pass an empty array to not update.
+   * @param origins the origins to add, pass an empty array to not update.
+   * @param dataCollectionPermissions the data collection permissions to add, pass an empty array to
+   *     not update.
+   * @return the updated {@link WebExtension} instance.
+   */
+  @NonNull
+  @AnyThread
+  public GeckoResult<WebExtension> addOptionalPermissions(
+      final @NonNull String extensionId,
+      @NonNull final String[] permissions,
+      @NonNull final String[] origins,
+      @NonNull final String[] dataCollectionPermissions) {
+    final GeckoBundle bundle = new GeckoBundle(4);
+    bundle.putString("extensionId", extensionId);
+    bundle.putStringArray("permissions", permissions);
+    bundle.putStringArray("origins", origins);
+    bundle.putStringArray("dataCollectionPermissions", dataCollectionPermissions);
+
+    return EventDispatcher.getInstance()
+        .queryBundle("GeckoView:WebExtension:AddOptionalPermissions", bundle)
+        .map(ext -> WebExtension.fromBundle(mDelegateControllerProvider, ext))
+        .map(this::registerWebExtension);
+  }
+
+  /**
+   * Remove the provided permissions from the {@link WebExtension} with the given id.
+   *
+   * @deprecated Use removeOptionalPermissions(WebExtension String[], String[], String[]) to account
+   *     for data collection permissions.
+   * @param extensionId the id of {@link WebExtension} instance to modify.
+   * @param permissions the permissions to remove, pass an empty array to not update.
+   * @param origins the origins to remove, pass an empty array to not update.
+   * @return the updated {@link WebExtension} instance.
+   */
+  @NonNull
+  @AnyThread
+  @Deprecated
+  @DeprecationSchedule(id = "web-extension-remove-optional-permissions", version = 143)
+  public GeckoResult<WebExtension> removeOptionalPermissions(
+      final @NonNull String extensionId,
+      @NonNull final String[] permissions,
+      @NonNull final String[] origins) {
+    return removeOptionalPermissions(extensionId, permissions, origins, new String[0]);
+  }
+
+  /**
+   * Remove the provided permissions from the {@link WebExtension} with the given id.
+   *
+   * @param extensionId the id of {@link WebExtension} instance to modify.
+   * @param permissions the permissions to remove, pass an empty array to not update.
+   * @param origins the origins to remove, pass an empty array to not update.
+   * @param dataCollectionPermissions the data collection permissions to remove, pass an array to
+   *     not update.
+   * @return the updated {@link WebExtension} instance.
+   */
+  @NonNull
+  @AnyThread
+  public GeckoResult<WebExtension> removeOptionalPermissions(
+      final @NonNull String extensionId,
+      @NonNull final String[] permissions,
+      @NonNull final String[] origins,
+      @NonNull final String[] dataCollectionPermissions) {
+    final GeckoBundle bundle = new GeckoBundle(4);
+    bundle.putString("extensionId", extensionId);
+    bundle.putStringArray("permissions", permissions);
+    bundle.putStringArray("origins", origins);
+    bundle.putStringArray("dataCollectionPermissions", dataCollectionPermissions);
+
+    return EventDispatcher.getInstance()
+        .queryBundle("GeckoView:WebExtension:RemoveOptionalPermissions", bundle)
+        .map(ext -> WebExtension.fromBundle(mDelegateControllerProvider, ext))
+        .map(this::registerWebExtension);
+  }
+
   /**
    * Install a built-in extension.
    *
@@ -839,7 +1115,7 @@ public class WebExtensionController {
    * @return A {@link GeckoResult} that will complete when the update process finishes. If an update
    *     is found and installed successfully, the GeckoResult will return the updated {@link
    *     WebExtension}. If no update is available, null will be returned. If the updated extension
-   *     requires new permissions, the {@link PromptDelegate#installPrompt} will be called.
+   *     requires new permissions, the {@link PromptDelegate#installPromptRequest} will be called.
    * @see PromptDelegate#updatePrompt
    */
   @AnyThread
@@ -883,7 +1159,7 @@ public class WebExtensionController {
     Log.d(LOGTAG, "handleMessage " + event);
 
     if ("GeckoView:WebExtension:InstallPrompt".equals(event)) {
-      installPrompt(bundle, callback);
+      installPromptRequest(bundle, callback);
       return;
     } else if ("GeckoView:WebExtension:UpdatePrompt".equals(event)) {
       updatePrompt(bundle, callback);
@@ -896,6 +1172,9 @@ public class WebExtensionController {
     } else if ("GeckoView:WebExtension:OnDisabling".equals(event)) {
       onDisabling(bundle);
       return;
+    } else if ("GeckoView:WebExtension:OnOptionalPermissionsChanged".equals(event)) {
+      onOptionalPermissionsChanged(bundle);
+      return;
     } else if ("GeckoView:WebExtension:OnDisabled".equals(event)) {
       onDisabled(bundle);
       return;
@@ -923,6 +1202,9 @@ public class WebExtensionController {
     } else if ("GeckoView:WebExtension:OnInstallationFailed".equals(event)) {
       onInstallationFailed(bundle);
       return;
+    } else if ("GeckoView:WebExtension:OnReady".equals(event)) {
+      onReady(bundle);
+      return;
     }
 
     extensionFromBundle(bundle)
@@ -1002,7 +1284,7 @@ public class WebExtensionController {
             });
   }
 
-  private void installPrompt(final GeckoBundle message, final EventCallback callback) {
+  private void installPromptRequest(final GeckoBundle message, final EventCallback callback) {
     final GeckoBundle extensionBundle = message.getBundle("extension");
     if (extensionBundle == null
         || !extensionBundle.containsKey("webExtensionId")
@@ -1023,39 +1305,46 @@ public class WebExtensionController {
       return;
     }
 
-    final GeckoResult<AllowOrDeny> promptResponse = mPromptDelegate.onInstallPrompt(extension);
+    // TODO - Bug 1970214: remove the first call when we remove the deprecated
+    // `onInstallPromptRequest` method since this has been done to preserve
+    // backward compatibility.
+    GeckoResult<WebExtension.PermissionPromptResponse> promptResponse =
+        mPromptDelegate.onInstallPromptRequest(
+            extension, message.getStringArray("permissions"), message.getStringArray("origins"));
+
+    if (promptResponse == null) {
+      promptResponse =
+          mPromptDelegate.onInstallPromptRequest(
+              extension,
+              message.getStringArray("permissions"),
+              message.getStringArray("origins"),
+              message.getStringArray("dataCollectionPermissions"));
+    }
+
     if (promptResponse == null) {
       return;
     }
 
     callback.resolveTo(
         promptResponse.map(
-            allowOrDeny -> {
-              final GeckoBundle response = new GeckoBundle(1);
-              response.putBoolean("allow", AllowOrDeny.ALLOW.equals(allowOrDeny));
+            userResponse -> {
+              final GeckoBundle response = new GeckoBundle(3);
+              response.putBoolean("allow", userResponse.isPermissionsGranted);
+              response.putBoolean("privateBrowsingAllowed", userResponse.isPrivateModeGranted);
+              response.putBoolean(
+                  "isTechnicalAndInteractionDataGranted",
+                  userResponse.isTechnicalAndInteractionDataGranted);
               return response;
             }));
   }
 
   private void updatePrompt(final GeckoBundle message, final EventCallback callback) {
-    final GeckoBundle currentBundle = message.getBundle("currentlyInstalled");
-    final GeckoBundle updatedBundle = message.getBundle("updatedExtension");
+    final WebExtension currentExtension =
+        new WebExtension(mDelegateControllerProvider, message.getBundle("currentlyInstalled"));
+    final WebExtension updatedExtension =
+        new WebExtension(mDelegateControllerProvider, message.getBundle("updatedExtension"));
     final String[] newPermissions = message.getStringArray("newPermissions");
     final String[] newOrigins = message.getStringArray("newOrigins");
-    if (currentBundle == null || updatedBundle == null) {
-      if (BuildConfig.DEBUG_BUILD) {
-        throw new RuntimeException("Missing bundle");
-      }
-
-      Log.e(LOGTAG, "Missing bundle");
-      return;
-    }
-
-    final WebExtension currentExtension =
-        new WebExtension(mDelegateControllerProvider, currentBundle);
-
-    final WebExtension updatedExtension =
-        new WebExtension(mDelegateControllerProvider, updatedBundle);
 
     if (mPromptDelegate == null) {
       Log.e(
@@ -1064,9 +1353,21 @@ public class WebExtensionController {
       return;
     }
 
-    final GeckoResult<AllowOrDeny> promptResponse =
+    // TODO - Bug 1974744: remove the first call when we remove the deprecated
+    // `onUpdatePrompt` method since this has been done to preserve backward
+    // compatibility.
+    GeckoResult<AllowOrDeny> promptResponse =
         mPromptDelegate.onUpdatePrompt(
             currentExtension, updatedExtension, newPermissions, newOrigins);
+
+    if (promptResponse == null) {
+      final String[] newDataCollectionPermissions =
+          message.getStringArray("newDataCollectionPermissions");
+      promptResponse =
+          mPromptDelegate.onUpdatePrompt(
+              updatedExtension, newPermissions, newOrigins, newDataCollectionPermissions);
+    }
+
     if (promptResponse == null) {
       return;
     }
@@ -1093,8 +1394,22 @@ public class WebExtensionController {
     final String[] permissions =
         message.bundle.getBundle("permissions").getStringArray("permissions");
     final String[] origins = message.bundle.getBundle("permissions").getStringArray("origins");
-    final GeckoResult<AllowOrDeny> promptResponse =
+
+    // TODO - Bug 1972510: remove the first call when we remove the deprecated
+    // `onOptionalPrompt` method since this has been done to preserve backward
+    // compatibility.
+    GeckoResult<AllowOrDeny> promptResponse =
         mPromptDelegate.onOptionalPrompt(extension, permissions, origins);
+
+    if (promptResponse == null) {
+      final String[] dataCollectionPermissions =
+          message.bundle.getBundle("permissions").getStringArray("data_collection");
+
+      promptResponse =
+          mPromptDelegate.onOptionalPrompt(
+              extension, permissions, origins, dataCollectionPermissions);
+    }
+
     if (promptResponse == null) {
       return;
     }
@@ -1117,13 +1432,15 @@ public class WebExtensionController {
     final int errorCode = bundle.getInt("error");
     final GeckoBundle extensionBundle = bundle.getBundle("extension");
     WebExtension extension = null;
+    final String extensionId = bundle.getString("addonId");
     final String extensionName = bundle.getString("addonName");
+    final String extensionVersion = bundle.getString("addonVersion");
 
     if (extensionBundle != null) {
       extension = new WebExtension(mDelegateControllerProvider, extensionBundle);
     }
     mAddonManagerDelegate.onInstallationFailed(
-        extension, new InstallException(errorCode, extensionName));
+        extension, new InstallException(errorCode, extensionId, extensionName, extensionVersion));
   }
 
   private void onDisabling(final GeckoBundle bundle) {
@@ -1137,6 +1454,17 @@ public class WebExtensionController {
     mAddonManagerDelegate.onDisabling(extension);
   }
 
+  private void onOptionalPermissionsChanged(final GeckoBundle bundle) {
+    if (mAddonManagerDelegate == null) {
+      Log.e(LOGTAG, "no AddonManager delegate registered");
+      return;
+    }
+
+    final GeckoBundle extensionBundle = bundle.getBundle("extension");
+    final WebExtension extension = new WebExtension(mDelegateControllerProvider, extensionBundle);
+    mAddonManagerDelegate.onOptionalPermissionsChanged(extension);
+  }
+
   private void onDisabled(final GeckoBundle bundle) {
     if (mAddonManagerDelegate == null) {
       Log.e(LOGTAG, "no AddonManager delegate registered");
@@ -1214,6 +1542,17 @@ public class WebExtensionController {
     mAddonManagerDelegate.onInstalled(extension);
   }
 
+  private void onReady(final GeckoBundle bundle) {
+    if (mAddonManagerDelegate == null) {
+      Log.e(LOGTAG, "no AddonManager delegate registered");
+      return;
+    }
+
+    final GeckoBundle extensionBundle = bundle.getBundle("extension");
+    final WebExtension extension = new WebExtension(mDelegateControllerProvider, extensionBundle);
+    mAddonManagerDelegate.onReady(extension);
+  }
+
   private void onDisabledProcessSpawning() {
     if (mExtensionProcessDelegate == null) {
       Log.e(LOGTAG, "no extension process delegate registered");
@@ -1386,7 +1725,7 @@ public class WebExtensionController {
     if (delegate != null) {
       result = delegate.onCloseTab(extension, message.session);
     } else {
-      result = GeckoResult.fromValue(AllowOrDeny.DENY);
+      result = GeckoResult.deny();
     }
 
     message.callback.resolveTo(
@@ -1513,11 +1852,7 @@ public class WebExtensionController {
     }
 
     private static boolean equals(final Object a, final Object b) {
-      if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT) {
-        return Objects.equals(a, b);
-      }
-
-      return (a == b) || (a != null && a.equals(b));
+      return Objects.equals(a, b);
     }
 
     @Override
